@@ -1,7 +1,9 @@
+pub mod adaptors;
+
 use crate::{
-    make_runtime_error, runtime_error,
+    runtime_error,
     value_iterator::{make_iterator, ValueIterator, ValueIteratorOutput as Output},
-    BinaryOp, DataMap, RuntimeResult, Value, ValueList, ValueMap, ValueVec, Vm,
+    BinaryOp, DataMap, RuntimeError, RuntimeResult, Value, ValueList, ValueMap, ValueVec, Vm,
 };
 
 pub fn make_module() -> ValueMap {
@@ -13,7 +15,6 @@ pub fn make_module() -> ValueMap {
         [iterable, f] if iterable.is_iterable() && f.is_callable() => {
             let iter = make_iterator(iterable).unwrap().map(collect_pair);
             let f = f.clone();
-            let vm = vm.child_vm();
 
             for output in iter {
                 match output {
@@ -45,7 +46,6 @@ pub fn make_module() -> ValueMap {
         [iterable, f] if iterable.is_iterable() && f.is_callable() => {
             let iter = make_iterator(iterable).unwrap().map(collect_pair);
             let f = f.clone();
-            let vm = vm.child_vm();
 
             for output in iter {
                 match output {
@@ -75,12 +75,12 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("chain", |vm, args| match vm.get_args(args) {
         [iterable_a, iterable_b] if iterable_a.is_iterable() && iterable_b.is_iterable() => {
-            let iter_a = make_iterator(iterable_a).unwrap();
-            let iter_b = make_iterator(iterable_b).unwrap();
+            let result = ValueIterator::make_external(adaptors::Chain::new(
+                make_iterator(iterable_a).unwrap(),
+                make_iterator(iterable_b).unwrap(),
+            ));
 
-            let mut iter = iter_a.chain(iter_b);
-
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            Ok(Iterator(result))
         }
         _ => runtime_error!("iterator.chain: Expected two iterables as arguments"),
     });
@@ -97,6 +97,11 @@ pub fn make_module() -> ValueMap {
         _ => runtime_error!("iterator.consume: Expected iterable as argument"),
     });
 
+    result.add_fn("copy", |vm, args| match vm.get_args(args) {
+        [Iterator(iter)] => Ok(Iterator(iter.make_copy())),
+        _ => runtime_error!("iterator.copy: Expected iterator as argument"),
+    });
+
     result.add_fn("count", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
             let mut result = 0;
@@ -113,36 +118,21 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("each", |vm, args| match vm.get_args(args) {
         [iterable, f] if iterable.is_iterable() && f.is_callable() => {
-            let iter = make_iterator(iterable).unwrap().map(collect_pair);
-            let f = f.clone();
-            let mut vm = vm.spawn_shared_vm();
+            let result = adaptors::Each::new(
+                make_iterator(iterable).unwrap(),
+                f.clone(),
+                vm.spawn_shared_vm(),
+            );
 
-            let mut iter = iter.map(move |iter_output| match iter_output {
-                Output::Value(value) => match vm.run_function(f.clone(), &[value]) {
-                    Ok(result) => Output::Value(result),
-                    Err(error) => Output::Error(error.with_prefix("iterator.each")),
-                },
-                Output::Error(error) => Output::Error(error),
-                _ => unreachable!(),
-            });
-
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            Ok(Iterator(ValueIterator::make_external(result)))
         }
         _ => runtime_error!("iterator.each: Expected iterable and function as arguments"),
     });
 
     result.add_fn("enumerate", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
-            let mut iter = make_iterator(iterable)
-                .unwrap()
-                .enumerate()
-                .map(|(i, iter_output)| match collect_pair(iter_output) {
-                    Output::Value(value) => Output::ValuePair(Number(i.into()), value),
-                    Output::Error(error) => Output::Error(error),
-                    _ => unreachable!(),
-                });
-
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            let result = adaptors::Enumerate::new(make_iterator(iterable).unwrap());
+            Ok(Iterator(ValueIterator::make_external(result)))
         }
         _ => runtime_error!("iterator.enumerate: Expected iterable as argument"),
     });
@@ -153,7 +143,6 @@ pub fn make_module() -> ValueMap {
                 let result = result.clone();
                 let f = f.clone();
                 let mut iter = make_iterator(iterable).unwrap();
-                let vm = vm.child_vm();
 
                 match iter
                     .lock_internals(|iterator| {
@@ -193,48 +182,17 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("intersperse", |vm, args| match vm.get_args(args) {
         [iterable, separator_fn] if iterable.is_iterable() && separator_fn.is_callable() => {
-            let mut iter = make_iterator(iterable).unwrap().peekable();
-            let mut intersperse = false;
-            let separator_fn = separator_fn.clone();
-            let mut vm = vm.spawn_shared_vm();
-
-            let result = move || {
-                if iter.peek().is_some() {
-                    let result = if intersperse {
-                        match vm.run_function(separator_fn.clone(), &[]) {
-                            Ok(result) => Output::Value(result),
-                            Err(error) => Output::Error(error.with_prefix("iterator.intersperse")),
-                        }
-                    } else {
-                        iter.next().unwrap()
-                    };
-                    intersperse = !intersperse;
-                    Some(result)
-                } else {
-                    None
-                }
-            };
+            let result = adaptors::IntersperseWith::new(
+                make_iterator(iterable).unwrap(),
+                separator_fn.clone(),
+                vm.spawn_shared_vm(),
+            );
 
             Ok(Iterator(ValueIterator::make_external(result)))
         }
         [iterable, separator] if iterable.is_iterable() => {
-            let mut iter = make_iterator(iterable).unwrap().peekable();
-            let mut intersperse = false;
-            let separator = separator.clone();
-
-            let result = move || {
-                if iter.peek().is_some() {
-                    let result = if intersperse {
-                        Output::Value(separator.clone())
-                    } else {
-                        iter.next().unwrap()
-                    };
-                    intersperse = !intersperse;
-                    Some(result)
-                } else {
-                    None
-                }
-            };
+            let result =
+                adaptors::Intersperse::new(make_iterator(iterable).unwrap(), separator.clone());
 
             Ok(Iterator(ValueIterator::make_external(result)))
         }
@@ -242,41 +200,13 @@ pub fn make_module() -> ValueMap {
     });
 
     result.add_fn("keep", |vm, args| match vm.get_args(args) {
-        [iterable, f] if iterable.is_iterable() && f.is_callable() => {
-            let mut iter = make_iterator(iterable).unwrap().map(collect_pair);
-            let f = f.clone();
-            let mut vm = vm.spawn_shared_vm();
-
-            Ok(Iterator(ValueIterator::make_external(move || {
-                for output in &mut iter {
-                    match output {
-                        Output::Value(value) => {
-                            match vm.run_function(f.clone(), &[value.clone()]) {
-                                Ok(Bool(result)) => {
-                                    if result {
-                                        return Some(Output::Value(value));
-                                    } else {
-                                        continue;
-                                    }
-                                }
-                                Ok(unexpected) => {
-                                    return Some(Output::Error(make_runtime_error!(format!(
-                                        "iterator.keep: Expected a Bool to be returned from the \
-                                         predicate, found '{}'",
-                                        unexpected.type_as_string(),
-                                    ))))
-                                }
-                                Err(error) => {
-                                    return Some(Output::Error(error.with_prefix("iterator.keep")))
-                                }
-                            }
-                        }
-                        error @ Output::Error(_) => return Some(error),
-                        _ => unreachable!(),
-                    }
-                }
-                None
-            })))
+        [iterable, predicate] if iterable.is_iterable() && predicate.is_callable() => {
+            let result = adaptors::Keep::new(
+                make_iterator(iterable).unwrap(),
+                predicate.clone(),
+                vm.spawn_shared_vm(),
+            );
+            Ok(Iterator(ValueIterator::make_external(result)))
         }
         _ => runtime_error!("iterator.keep: Expected iterable and function as arguments"),
     });
@@ -302,38 +232,14 @@ pub fn make_module() -> ValueMap {
     result.add_fn("max", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
             let iterable = iterable.clone();
-            let vm = vm.child_vm();
-            let mut result: Option<Value> = None;
-
-            for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
-                match iter_output {
-                    Output::Value(value) => {
-                        result = Some(match result {
-                            Some(result) => match vm.run_binary_op(
-                                BinaryOp::Less,
-                                result.clone(),
-                                value.clone(),
-                            ) {
-                                Ok(Bool(true)) => value,
-                                Ok(Bool(false)) => result,
-                                Ok(unexpected) => {
-                                    return runtime_error!(
-                                        "iterator.max: \
-                                         Expected Bool from < comparison, found '{}'",
-                                        unexpected.type_as_string()
-                                    );
-                                }
-                                Err(error) => return Err(error.with_prefix("iterator.max")),
-                            },
-                            None => value,
-                        })
-                    }
-                    Output::Error(error) => return Err(error.with_prefix("iterator.max")),
-                    _ => unreachable!(),
-                }
-            }
-
-            Ok(result.unwrap_or(Empty))
+            run_iterator_comparison(vm, iterable, InvertResult::Yes)
+                .map_err(|e| e.with_prefix("iterator.max"))
+        }
+        [iterable, key_fn] if iterable.is_iterable() && key_fn.is_callable() => {
+            let iterable = iterable.clone();
+            let key_fn = key_fn.clone();
+            run_iterator_comparison_by_key(vm, iterable, key_fn, InvertResult::Yes)
+                .map_err(|e| e.with_prefix("iterator.max"))
         }
         _ => runtime_error!("iterator.max: Expected iterable as argument"),
     });
@@ -341,38 +247,14 @@ pub fn make_module() -> ValueMap {
     result.add_fn("min", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
             let iterable = iterable.clone();
-            let vm = vm.child_vm();
-            let mut result: Option<Value> = None;
-
-            for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
-                match iter_output {
-                    Output::Value(value) => {
-                        result = Some(match result {
-                            Some(result) => match vm.run_binary_op(
-                                BinaryOp::Less,
-                                result.clone(),
-                                value.clone(),
-                            ) {
-                                Ok(Bool(true)) => result,
-                                Ok(Bool(false)) => value,
-                                Ok(unexpected) => {
-                                    return runtime_error!(
-                                        "iterator.min: \
-                                         Expected Bool from < comparison, found '{}'",
-                                        unexpected.type_as_string()
-                                    );
-                                }
-                                Err(error) => return Err(error.with_prefix("iterator.min")),
-                            },
-                            None => value,
-                        })
-                    }
-                    Output::Error(error) => return Err(error.with_prefix("iterator.min")),
-                    _ => unreachable!(),
-                }
-            }
-
-            Ok(result.unwrap_or(Empty))
+            run_iterator_comparison(vm, iterable, InvertResult::No)
+                .map_err(|e| e.with_prefix("iterator.min"))
+        }
+        [iterable, key_fn] if iterable.is_iterable() && key_fn.is_callable() => {
+            let iterable = iterable.clone();
+            let key_fn = key_fn.clone();
+            run_iterator_comparison_by_key(vm, iterable, key_fn, InvertResult::No)
+                .map_err(|e| e.with_prefix("iterator.min"))
         }
         _ => runtime_error!("iterator.min: Expected iterable as argument"),
     });
@@ -380,32 +262,17 @@ pub fn make_module() -> ValueMap {
     result.add_fn("min_max", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
             let iterable = iterable.clone();
-            let vm = vm.child_vm();
             let mut result = None;
 
-            let compare_values = |vm: &mut Vm, op, a: Value, b: Value| -> RuntimeResult {
-                match vm.run_binary_op(op, a.clone(), b.clone()) {
-                    Ok(Bool(true)) => Ok(a),
-                    Ok(Bool(false)) => Ok(b),
-                    Ok(unexpected) => {
-                        return runtime_error!(
-                            "iterator.min_max: \
-                             Expected Bool from {} comparison, found '{}'",
-                            op,
-                            unexpected.type_as_string()
-                        );
-                    }
-                    Err(error) => Err(error.with_prefix("iterator.min_max")),
-                }
-            };
-
             for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
                 match iter_output {
                     Output::Value(value) => {
                         result = Some(match result {
                             Some((min, max)) => (
-                                compare_values(vm, BinaryOp::Less, min, value.clone())?,
-                                compare_values(vm, BinaryOp::Greater, max, value)?,
+                                compare_values(vm, min, value.clone(), InvertResult::No)
+                                    .map_err(|e| e.with_prefix("iterator.min_max"))?,
+                                compare_values(vm, max, value, InvertResult::Yes)
+                                    .map_err(|e| e.with_prefix("iterator.min_max"))?,
                             ),
                             None => (value.clone(), value),
                         })
@@ -417,6 +284,44 @@ pub fn make_module() -> ValueMap {
 
             Ok(result.map_or(Empty, |(min, max)| Tuple(vec![min, max].into())))
         }
+        [iterable, key_fn] if iterable.is_iterable() && key_fn.is_callable() => {
+            let iterable = iterable.clone();
+            let key_fn = key_fn.clone();
+            let mut result = None;
+
+            for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
+                match iter_output {
+                    Output::Value(value) => {
+                        let key = vm.run_function(key_fn.clone(), &[value.clone()])?;
+                        let value_and_key = (value, key);
+
+                        result = Some(match result {
+                            Some((min_and_key, max_and_key)) => (
+                                compare_values_with_key(
+                                    vm,
+                                    min_and_key,
+                                    value_and_key.clone(),
+                                    InvertResult::No,
+                                )
+                                .map_err(|e| e.with_prefix("iterator.min_max"))?,
+                                compare_values_with_key(
+                                    vm,
+                                    max_and_key,
+                                    value_and_key,
+                                    InvertResult::Yes,
+                                )
+                                .map_err(|e| e.with_prefix("iterator.min_max"))?,
+                            ),
+                            None => (value_and_key.clone(), value_and_key),
+                        })
+                    }
+                    Output::Error(error) => return Err(error),
+                    _ => unreachable!(),
+                }
+            }
+
+            Ok(result.map_or(Empty, |((min, _), (max, _))| Tuple(vec![min, max].into())))
+        }
         _ => runtime_error!("iterator.min_max: Expected iterable as argument"),
     });
 
@@ -434,7 +339,6 @@ pub fn make_module() -> ValueMap {
         [iterable, f] if iterable.is_iterable() && f.is_callable() => {
             let iter = make_iterator(iterable).unwrap().map(collect_pair);
             let f = f.clone();
-            let vm = vm.child_vm();
 
             for (i, output) in iter.enumerate() {
                 match output {
@@ -486,7 +390,7 @@ pub fn make_module() -> ValueMap {
                 }
             }
 
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            Ok(Iterator(iter))
         }
         _ => {
             runtime_error!("iterator.skip: Expected iterable and non-negative number as arguments")
@@ -508,9 +412,8 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("take", |vm, args| match vm.get_args(args) {
         [iterable, Number(n)] if iterable.is_iterable() && *n >= 0.0 => {
-            let mut iter = make_iterator(iterable).unwrap().take(n.into());
-
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            let result = adaptors::Take::new(make_iterator(iterable).unwrap(), n.into());
+            Ok(Iterator(ValueIterator::make_external(result)))
         }
         _ => {
             runtime_error!("iterator.take: Expected iterable and non-negative number as arguments")
@@ -519,9 +422,11 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("to_list", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
-            let mut result = ValueVec::new();
+            let iterator = make_iterator(iterable).unwrap();
+            let (size_hint, _) = iterator.size_hint();
+            let mut result = ValueVec::with_capacity(size_hint);
 
-            for output in make_iterator(iterable).unwrap().map(collect_pair) {
+            for output in iterator.map(collect_pair) {
                 match output {
                     Output::Value(value) => result.push(value),
                     Output::Error(error) => return Err(error),
@@ -536,10 +441,15 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("to_map", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
-            let mut result = DataMap::new();
+            let iterator = make_iterator(iterable).unwrap();
+            let (size_hint, _) = iterator.size_hint();
+            let mut result = DataMap::with_capacity(size_hint);
 
-            for output in make_iterator(iterable).unwrap() {
+            for output in iterator {
                 match output {
+                    Output::ValuePair(key, value) => {
+                        result.insert(key.into(), value);
+                    }
                     Output::Value(Tuple(t)) if t.data().len() == 2 => {
                         let key = t.data()[0].clone();
                         let value = t.data()[1].clone();
@@ -548,9 +458,6 @@ pub fn make_module() -> ValueMap {
                     Output::Value(value) => {
                         result.insert(value.into(), Value::Empty);
                     }
-                    Output::ValuePair(key, value) => {
-                        result.insert(key.into(), value);
-                    }
                     Output::Error(error) => return Err(error),
                 }
             }
@@ -562,9 +469,11 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("to_string", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
-            let mut result = String::new();
+            let iterator = make_iterator(iterable).unwrap();
+            let (size_hint, _) = iterator.size_hint();
+            let mut result = String::with_capacity(size_hint);
 
-            for output in make_iterator(iterable).unwrap().map(collect_pair) {
+            for output in iterator.map(collect_pair) {
                 match output {
                     Output::Value(Str(s)) => result.push_str(&s),
                     Output::Value(value) => result.push_str(&value.to_string()),
@@ -580,9 +489,11 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("to_tuple", |vm, args| match vm.get_args(args) {
         [iterable] if iterable.is_iterable() => {
-            let mut result = Vec::new();
+            let iterator = make_iterator(iterable).unwrap();
+            let (size_hint, _) = iterator.size_hint();
+            let mut result = Vec::with_capacity(size_hint);
 
-            for output in make_iterator(iterable).unwrap().map(collect_pair) {
+            for output in iterator.map(collect_pair) {
                 match output {
                     Output::Value(value) => result.push(value),
                     Output::Error(error) => return Err(error),
@@ -597,18 +508,11 @@ pub fn make_module() -> ValueMap {
 
     result.add_fn("zip", |vm, args| match vm.get_args(args) {
         [iterable_a, iterable_b] if iterable_a.is_iterable() && iterable_b.is_iterable() => {
-            let iter_a = make_iterator(iterable_a).unwrap().map(collect_pair);
-            let iter_b = make_iterator(iterable_b).unwrap().map(collect_pair);
-
-            let mut iter = iter_a.zip(iter_b).map(|(a, b)| match (a, b) {
-                (Output::Value(output_a), Output::Value(output_b)) => {
-                    Output::ValuePair(output_a, output_b)
-                }
-                (Output::Error(e), _) | (_, Output::Error(e)) => Output::Error(e),
-                _ => unreachable!(),
-            });
-
-            Ok(Iterator(ValueIterator::make_external(move || iter.next())))
+            let result = adaptors::Zip::new(
+                make_iterator(iterable_a).unwrap(),
+                make_iterator(iterable_b).unwrap(),
+            );
+            Ok(Iterator(ValueIterator::make_external(result)))
         }
         _ => runtime_error!("iterator.zip: Expected two iterables as arguments"),
     });
@@ -629,7 +533,6 @@ fn fold_with_operator(
     initial_value: Value,
     operator: BinaryOp,
 ) -> RuntimeResult {
-    let vm = vm.child_vm();
     let mut result = initial_value;
 
     for output in make_iterator(&iterable).unwrap().map(collect_pair) {
@@ -644,3 +547,109 @@ fn fold_with_operator(
 
     Ok(result)
 }
+
+fn run_iterator_comparison(
+    vm: &mut Vm,
+    iterable: Value,
+    invert_result: InvertResult,
+) -> RuntimeResult {
+    let mut result: Option<Value> = None;
+
+    for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
+        match iter_output {
+            Output::Value(value) => {
+                result = Some(match result {
+                    Some(result) => {
+                        compare_values(vm, result.clone(), value.clone(), invert_result)?
+                    }
+                    None => value,
+                })
+            }
+            Output::Error(error) => return Err(error),
+            _ => unreachable!(),
+        }
+    }
+
+    Ok(result.unwrap_or_default())
+}
+
+fn run_iterator_comparison_by_key(
+    vm: &mut Vm,
+    iterable: Value,
+    key_fn: Value,
+    invert_result: InvertResult,
+) -> RuntimeResult {
+    let mut result_and_key: Option<(Value, Value)> = None;
+
+    for iter_output in make_iterator(&iterable).unwrap().map(collect_pair) {
+        match iter_output {
+            Output::Value(value) => {
+                let key = vm.run_function(key_fn.clone(), &[value.clone()])?;
+                let value_and_key = (value, key);
+
+                result_and_key = Some(match result_and_key {
+                    Some(result_and_key) => {
+                        compare_values_with_key(vm, result_and_key, value_and_key, invert_result)?
+                    }
+                    None => value_and_key,
+                });
+            }
+            Output::Error(error) => return Err(error),
+            _ => unreachable!(),
+        }
+    }
+
+    Ok(result_and_key.map_or(Value::Empty, |(value, _)| value))
+}
+
+// Compares two values using BinaryOp::Less
+//
+// Returns the lesser of the two values, unless `invert_result` is set to Yes
+fn compare_values(vm: &mut Vm, a: Value, b: Value, invert_result: InvertResult) -> RuntimeResult {
+    use {InvertResult::*, Value::Bool};
+
+    let comparison_result = vm.run_binary_op(BinaryOp::Less, a.clone(), b.clone())?;
+
+    match (comparison_result, invert_result) {
+        (Bool(true), No) => Ok(a),
+        (Bool(false), No) => Ok(b),
+        (Bool(true), Yes) => Ok(b),
+        (Bool(false), Yes) => Ok(a),
+        (other, _) => runtime_error!(
+            "Expected Bool from '<' comparison, found '{}'",
+            other.type_as_string()
+        ),
+    }
+}
+
+// Compares two values using BinaryOp::Less
+//
+// Returns the lesser of the two values, unless `invert_result` is set to Yes
+fn compare_values_with_key(
+    vm: &mut Vm,
+    a_and_key: (Value, Value),
+    b_and_key: (Value, Value),
+    invert_result: InvertResult,
+) -> Result<(Value, Value), RuntimeError> {
+    use {InvertResult::*, Value::Bool};
+
+    let comparison_result =
+        vm.run_binary_op(BinaryOp::Less, a_and_key.1.clone(), b_and_key.1.clone())?;
+
+    match (comparison_result, invert_result) {
+        (Bool(true), No) => Ok(a_and_key),
+        (Bool(false), No) => Ok(b_and_key),
+        (Bool(true), Yes) => Ok(b_and_key),
+        (Bool(false), Yes) => Ok(a_and_key),
+        (other, _) => runtime_error!(
+            "Expected Bool from '<' comparison, found '{}'",
+            other.type_as_string()
+        ),
+    }
+}
+
+#[derive(Clone, Copy)]
+enum InvertResult {
+    Yes,
+    No,
+}
