@@ -174,11 +174,6 @@ a = 1
             test_script(script, number_list(&[1, 1, 1]));
         }
 
-        #[test]
-        fn from_range() {
-            test_script("[3..0]", number_list(&[3, 2, 1]));
-        }
-
         #[test]
         fn access_element() {
             let script = "
@@ -574,7 +569,7 @@ match x
         #[test]
         fn match_list_subslice() {
             let script = "
-x = [1..=5]
+x = (1..=5).to_list()
 match x
   [0, ...] then 0
   [..., 1] then -1
@@ -587,7 +582,7 @@ match x
         #[test]
         fn match_list_subslice_with_id() {
             let script = "
-x = [1..=5]
+x = (1..=5).to_list()
 match x
   [0, rest...] then rest
   [rest..., 3, 2, 1] then rest
@@ -600,7 +595,7 @@ match x
         #[test]
         fn match_list_subslice_at_start_with_id() {
             let script = "
-x = [1..=5]
+x = (1..=5).to_list()
 match x
   [0, rest...] then rest
   [rest..., 3, 4, 5] then rest
@@ -797,6 +792,24 @@ add(5, 6)";
             test_script(script, Number(11.0.into()));
         }
 
+        #[test]
+        fn call_with_insufficient_args() {
+            let script = "
+foo = |a, b| b
+foo 42
+";
+            test_script(script, Empty);
+        }
+
+        #[test]
+        fn call_with_extra_args() {
+            let script = "
+foo = |a, b| a + b
+foo 10, 20, 30, 40
+";
+            test_script(script, Number(30.into()));
+        }
+
         #[test]
         fn nested_call_without_parens() {
             let script = "
@@ -897,6 +910,14 @@ f 5, 10, 20, 30";
             test_script(script, Number(65.0.into()));
         }
 
+        #[test]
+        fn variadic_function_with_missing_args() {
+            let script = "
+f = |a, b...| b
+f()";
+            test_script(script, Empty);
+        }
+
         #[test]
         fn nested_function() {
             let script = "
@@ -1300,7 +1321,7 @@ sum
             result_data.add_value("bar", Str("baz".into()));
 
             test_script(
-                "{foo: 42, bar: \"baz\"}",
+                "{foo: 42, bar: 'baz'}",
                 Map(ValueMap::with_data(result_data)),
             );
         }
@@ -1340,6 +1361,15 @@ m.baz";
             test_script(script, Number(f64::from(-1).into()));
         }
 
+        #[test]
+        fn string_keys() {
+            let script = r#"
+foo, bar = 42, -1
+m = {foo, bar, 'baz': 99}
+m.baz"#;
+            test_script(script, Number(99.into()));
+        }
+
         #[test]
         fn instance_function_no_args() {
             let script = "
@@ -1553,35 +1583,35 @@ m.foo
         fn function_body_in_iterator_chain() {
             // The result.insert() call in a function block, followed by a continued iterator chain
             // at a lower indentation level caused a parser error.
-            let script = r#"
+            let script = "
 result = {}
 (1..=5)
   .each |x|
     result.insert(x, x * x)
   .consume()
 result.size()
-"#;
+";
             test_script(script, Number(5.0.into()));
         }
 
         #[test]
         fn inline_function_body_in_call_args() {
-            let script = r#"
+            let script = "
 equal = |x, y| x == y
 equal
   (0..10).position(|n| n == 5),
   5
-"#;
+";
             test_script(script, Bool(true));
         }
 
         #[test]
         fn range_in_call_args() {
-            let script = r#"
+            let script = "
 foo = |range, x| range.size() + x
 min, max = 0, 10
 foo min..max, 20
-"#;
+";
             test_script(script, Number(30.0.into()));
         }
     }
@@ -1646,6 +1676,17 @@ gen(1..=5).to_tuple()";
             test_script(script, number_tuple(&[1, 2, 3, 4, 5]));
         }
 
+        #[test]
+        fn generator_with_missing_arg() {
+            let script = "
+gen = |xs|
+  xs = if xs == () then (1, 2, 3) else xs
+  for x in xs
+    yield x
+gen().to_tuple()";
+            test_script(script, number_tuple(&[1, 2, 3]));
+        }
+
         #[test]
         fn generator_variadic() {
             let script = "
@@ -1678,9 +1719,22 @@ gen().to_tuple()
 ";
             test_script(script, number_tuple(&[1, 2, 3]));
         }
+
+        #[test]
+        fn generator_with_captured_data_and_missing_args() {
+            let script = "
+x = 1, 2, 3
+gen = |offset, bar...|
+  offset = if offset == () then 10 else offset
+  for y in x
+    yield y + offset
+gen().to_tuple()
+";
+            test_script(script, number_tuple(&[11, 12, 13]));
+        }
     }
 
-    mod num2_test {
+    mod num2 {
         use super::*;
 
         #[test]
@@ -1738,9 +1792,44 @@ x = num2 4, 5
 x[1]";
             test_script(script, Number(5.0.into()));
         }
+
+        #[test]
+        fn index_element_mutation() {
+            let script = "
+x = num2 4, 5
+x[1] = 99
+x[1]";
+            test_script(script, Number(99.0.into()));
+        }
+
+        #[test]
+        fn index_range_mutation() {
+            let script = "
+x = num2 4, 5
+x[0..2] = 1
+x[0] + x[1]";
+            test_script(script, Number(2.0.into()));
+        }
+
+        #[test]
+        fn index_full_range_mutation() {
+            let script = "
+x = num2 4, 5
+x[..] = 2
+x";
+            test_script(script, num2(2.0, 2.0));
+        }
+
+        #[test]
+        fn iterator_ops() {
+            let script = "
+num2(1, -1).keep(|n| n < 0).count()
+";
+            test_script(script, Number(1.0.into()));
+        }
     }
 
-    mod num4_test {
+    mod num4 {
         use super::*;
 
         #[test]
@@ -1819,6 +1908,41 @@ x = num4 9, 8, 7, 6
 x[3]";
             test_script(script, Number(6.0.into()));
         }
+
+        #[test]
+        fn index_element_mutation() {
+            let script = "
+x = num4 4, 5, 6, 7
+x[2] = 99
+x[2]";
+            test_script(script, Number(99.0.into()));
+        }
+
+        #[test]
+        fn index_range_mutation() {
+            let script = "
+x = num4 4, 5, 6, 7
+x[1..=2] = 1
+x";
+            test_script(script, num4(4.0, 1.0, 1.0, 7.0));
+        }
+
+        #[test]
+        fn index_full_range_mutation() {
+            let script = "
+x = num4 4, 5, 6, 7
+x[..] = 2
+x.sum()";
+            test_script(script, Number(8.0.into()));
+        }
+
+        #[test]
+        fn iterator_ops() {
+            let script = "
+num4(1, -1, 2, -2).keep(|n| n > 0).count()
+";
+            test_script(script, Number(2.0.into()));
+        }
     }
 
     mod strings {
@@ -1877,6 +2001,14 @@ x[3]";
             test_script("'héllö'[..=2]", string("hél"));
         }
 
+        #[test]
+        fn index_from_one_past_the_end() {
+            test_script("'x'[0..1]", string("x"));
+            test_script("'x'[1..]", string(""));
+            test_script("'x'[1..1]", string(""));
+            test_script("'héllö'[5..]", string(""));
+        }
+
         #[test]
         fn index_whole_string() {
             test_script("'héllö'[..]", string("héllö"));
@@ -1887,6 +2019,156 @@ x[3]";
             test_script("'héllö'[3..][..]", string("lö"));
             test_script("'héllö'[3..][1]", string("ö"));
         }
+
+        #[test]
+        fn escaped_backslash() {
+            test_script(r#""\\""#, string("\\"));
+        }
+
+        #[test]
+        fn interpolated_id() {
+            let script = "
+x = 1
+'$x + $x'
+";
+            test_script(script, string("1 + 1"));
+        }
+
+        #[test]
+        fn interpolated_id_from_capture() {
+            let script = "
+x = 1
+f = || '$x.$x'
+f()
+";
+            test_script(script, string("1.1"));
+        }
+
+        #[test]
+        fn interpolated_expression() {
+            let script = "
+x = 100
+'sqrt(x): ${x.sqrt()}'
+";
+            test_script(script, string("sqrt(x): 10.0"));
+        }
+
+        #[test]
+        fn interpolated_expression_nested() {
+            let script = "
+'foo${': ${42}'}'
+";
+            test_script(script, string("foo: 42"));
+        }
+
+        #[test]
+        fn interpolated_expression_inline_map() {
+            let script = "
+foo = |m| m.size()
+'${foo {bar: 42, baz: 99}}!'
+";
+            test_script(script, string("2!"));
+        }
+
+        #[test]
+        fn interpolated_expression_using_capture() {
+            let script = "
+x = 10
+f = || 'x * 2 == ${x * 2}'
+f()
+";
+            test_script(script, string("x * 2 == 20"));
+        }
+
+        #[test]
+        fn interpolated_string_as_map_key() {
+            let script = "
+x = 99
+m =
+  'key$x': 'foo'
+m.key99
+";
+            test_script(script, string("foo"));
+        }
+
+        #[test]
+        fn interpolated_string_in_lookup() {
+            let script = "
+x = 99
+m =
+  'key$x': 'foo'
+m.'key$x'
+";
+            test_script(script, string("foo"));
+        }
+
+        #[test]
+        fn interpolated_string_in_lookup_assignment() {
+            let script = "
+x = 99
+m =
+  'key$x': 'foo'
+m.'key$x' = 123
+m.'key$x'
+";
+            test_script(script, Number(123.into()));
+        }
+
+        #[test]
+        fn interpolated_string_with_value_with_overloaded_display() {
+            let script = "
+foo = {@display: |self| 'Foo'}
+'$foo'
+";
+            test_script(script, string("Foo"));
+        }
+    }
+
+    mod iterators {
+        use super::*;
+
+        #[test]
+        fn iterator_copy() {
+            let script = "
+x = (1..10).iter()
+z = x.copy()
+x.next()
+x.next()
+z.next()
+z.next()
+";
+            test_script(script, Number(2.into()));
+        }
+
+        #[test]
+        fn iterators_in_a_deep_copy() {
+            let script = "
+r = 1..10
+x = [r.iter()]
+z = x.deep_copy()
+x[0].next()
+x[0].next()
+z[0].next()
+z[0].next()
+";
+            test_script(script, Number(2.into()));
+        }
+
+        #[test]
+        fn copy_of_a_generator() {
+            let script = "
+generator = ||
+  for x in (1, 2, 3, 4, 5)
+    yield x
+x = generator()
+x.next() # 1
+y = x.copy()
+x.next() # 2
+x.next() # 3
+y.next()
+";
+            test_script(script, Number(2.into()));
+        }
     }
 
     mod error_recovery {
@@ -1902,7 +2184,7 @@ try
 catch _
   x + 1
 ";
-            test_script(script, Number(3.0.into()));
+            test_script(script, Number(3.into()));
         }
 
         #[test]
@@ -1930,7 +2212,7 @@ try
 catch error
   error.data
 "#;
-            test_script(script, Number(2.0.into()));
+            test_script(script, Number(2.into()));
         }
 
         #[test]
@@ -1943,7 +2225,7 @@ catch e
 finally
   99
 ";
-            test_script(script, Number(99.0.into()));
+            test_script(script, Number(99.into()));
         }
 
         #[test]
@@ -1961,7 +2243,7 @@ try
 catch _
   x += 1
 ";
-            test_script(script, Number(4.0.into()));
+            test_script(script, Number(4.into()));
         }
     }
 
@@ -1983,7 +2265,7 @@ locals.foo_meta =
 z = ((foo 2) * (foo 10) / (foo 4) + (foo 1) - (foo 2)) % foo 3
 z.x
 ";
-            test_script(script, Number(1.0.into()));
+            test_script(script, Number(1.into()));
         }
 
         #[test]
